@@ -1,16 +1,19 @@
-
-
 import os
 from os import path
+import json
+import pickle
 import operator as op
 import warnings
+from ase import constraints
 import numpy as np
 from ase.atoms import Atoms
 from ase.calculators.calculator import kpts2ndarray, kpts2sizeandoffsets
 from ase.constraints import FixAtoms, FixCartesian
 from ase.data import chemical_symbols, atomic_numbers
-from ase.io.espresso import construct_namelist, grep_valence, SSSP_VALENCE
+from ase.io.espresso import read_espresso_in, construct_namelist, grep_valence, SSSP_VALENCE, read_fortran_namelist
 from ase.units import create_units
+from pprint import pprint
+from xespresso.utils import check_type
 
 
 def write_espresso_in(filename, atoms, input_data={}, pseudopotentials=None,
@@ -66,12 +69,16 @@ def build_section_str(atoms, species_info, input_data, input_parameters):
     input_parameters['system']['nat'] = len(atoms)
 
     #
-    if 'input_ntyp' in input_data:
-        for key, value in input_data['input_ntyp'].items():
+    if 'INPUT_NTYP' in input_data:
+        for key, value in input_data['INPUT_NTYP'].items():
             for species in value:
                 if species in species_info:
                     mag_str = '{0}({1})'.format(key, species_info[species]['index'])
                     input_parameters['system'][mag_str] = value[species]
+    if 'hubbard_v' in input_data:
+        for key, value in input_data['hubbard_v'].items():
+                    mag_str = 'Hubbard_V{0}'.format(key)
+                    input_parameters['system'][mag_str] = value
     # Use cell as given or fit to a specific ibrav
     if 'ibrav' in input_parameters['system']:
         ibrav = input_parameters['system']['ibrav']
@@ -117,9 +124,9 @@ def build_atomic_species_str(atoms, input_parameters, pseudopotentials):
 
     # Species info holds the information on the pseudopotential and
     # associated for each element
-    if 'species' not in atoms.arrays:
+    if 'species' not in atoms.info:
         # print('no species')
-        atoms.arrays['species'] = atoms.get_chemical_symbols()
+        atoms.info['species'] = atoms.get_chemical_symbols()
     if pseudopotentials is None:
         pseudopotentials = {}
     species_info = {}
@@ -127,7 +134,7 @@ def build_atomic_species_str(atoms, input_parameters, pseudopotentials):
     ntyp = 0
     # Convert atoms into species.
     for i in range(len(atoms)):
-        species = atoms.arrays['species'][i]
+        species = atoms.info['species'][i]
         if species not in species_info:
             ntyp += 1
             species_info[species] = {}
@@ -201,7 +208,6 @@ def build_kpts_str(atoms, kspacing, kpts, koffset):
             kpts_str.append('{k[0]:.14f} {k[1]:.14f} {k[2]:.14f} 0\n'.format(k=k))
         kpts_str.append('\n')
     elif isinstance(kgrid, str) and (kgrid == "gamma"):
-        pwi.append()
         kpts_str = ['K_POINTS gamma\n']
         kpts_str.append('\n')
     else:
@@ -215,8 +221,12 @@ def build_atomic_positions_str(atoms, crystal_coordinates):
     '''
 
     '''
-    if 'species' not in atoms.arrays:
-        atoms.arrays['species'] = atoms.get_chemical_symbols()
+    import sys
+    if 'species' not in atoms.info:
+        atoms.info['species'] = atoms.get_chemical_symbols()
+    else:
+        if len(atoms.info['species']) != len(atoms):
+            sys.exit('Species is wrong, please check!')
     # Convert ase constraints to QE constraints
     # Nx3 array of force multipliers matches what QE uses
     # Do this early so it is available when constructing the atoms card
@@ -235,7 +245,7 @@ def build_atomic_positions_str(atoms, crystal_coordinates):
         atomic_positions_str = ['ATOMIC_POSITIONS angstrom\n']
     for i in range(len(atoms)):
         atom = atoms[i]
-        species = atoms.arrays['species'][i]
+        species = atoms.info['species'][i]
         # only inclued mask if something is fixed
         if not all(constraint_mask[atom.index]):
             mask = ' {mask[0]} {mask[1]} {mask[2]}'.format(
@@ -252,9 +262,10 @@ def build_atomic_positions_str(atoms, crystal_coordinates):
             '{mask}\n'.format(species=species, coords=coords, mask=mask))
     return atomic_positions_str
 
-def write_neb_in(filename, images, climbing_images, path_data = None, input_data=None, pseudopotentials=None,
-                      kspacing=None, kpts=None, koffset=(0, 0, 0),
-                      crystal_coordinates=False, **kwargs):
+def write_neb_in(filename, images, climbing_images = [], path_data = {}, 
+                 input_data={}, pseudopotentials=None,
+                 kspacing=None, kpts=None, koffset=(0, 0, 0),
+                 crystal_coordinates=False, **kwargs):
     """
     Create an input file for neb.x.
 
@@ -328,6 +339,7 @@ def write_neb_in(filename, images, climbing_images, path_data = None, input_data
 def get_atomic_species(pwo):
     '''
     '''
+    atomic_species = []
     with open(pwo) as f:
         lines = f.readlines()
         for idx, line in enumerate(lines):
@@ -337,4 +349,157 @@ def get_atomic_species(pwo):
                 atomic_species = [at_line.split()[1]
                     for at_line in lines[idx + 1:idx + 1 + nat]]
                 break
-    return atomic_species
\ No newline at end of file
+    return atomic_species
+
+def read_espresso_input(fileobj, neb = False):
+    """Parse a Quantum ESPRESSO input files, '.in', '.pwi'.
+    """
+    atoms = read_espresso_in(fileobj)
+    if isinstance(fileobj, str):
+        fileobj = open(fileobj, 'rU')
+    # parse namelist section and extract remaining lines
+    data, card_lines = read_fortran_namelist(fileobj)
+    input_data = {}
+    for sec, paras in data.items():
+        for key, value in paras.items():
+            input_data[key] = value
+    # get number of type
+    ntyp = data['system']['ntyp']
+    pseudopotentials = {}
+    trimmed_lines = (line for line in card_lines
+                     if line.strip() and not line[0] == '#')
+    for line in trimmed_lines:
+        if line.strip().startswith('ATOMIC_SPECIES'):
+            for i in range(ntyp):
+                species = next(trimmed_lines).split()
+                pseudopotentials[species[0]] = species[2]
+        if line.strip().startswith('K_POINTS'):
+            kpts = next(trimmed_lines)
+    # calc = Espresso(pseudopotentials=pseudopotentials, 
+    #                 input_data = input_data,
+    #                 kpts=kpts)
+    return atoms, input_data, pseudopotentials, kpts, constraints
+
+def sort_qe_input(parameters, package = 'PW'):
+    """
+    """
+    from xespresso.input_parameters import qe_namespace, default_parameters
+    import copy
+    pw_parameters = qe_namespace[package]
+    if 'input_data' not in parameters:
+        parameters['input_data'] = {}
+    sorted_parameters = copy.deepcopy(parameters)
+    unuse_parameters = {}
+    # section_names = ['CONTROL', 'SYSTEM', 'ELECTRONS', 'IONS', 'CELL', 'ATOMIC_SPECIES', 'K_POINTS', 'CELL_PARAMETERS', 'CONSTRAINTS', 'OCCUPATIONS', 'ATOMIC_VELECITIES', 'ATOMIC_FORCES']
+    section_names = ['CONTROL', 'SYSTEM', 'ELECTRONS', 'IONS', 'CELL', 'INPUT_NTYP']
+    for section in section_names:
+        if section not in sorted_parameters['input_data']: sorted_parameters['input_data'][section] = {}
+    for key, value in parameters.items():
+        if key in ['pseudopotentials', 'kpts', 'kspacing', 'koffset', 'input_data', 'climbing_images', 'path_data', 'crystal_coordinates']:
+            continue
+        flag = False
+        for section in section_names:
+            if key.upper() == section and isinstance(value, dict):
+                sorted_parameters['input_data'][section].update(value)
+                flag = True
+                del sorted_parameters[key]
+                break
+            if section.upper() == 'INPUT_NTYP': continue
+            if key in pw_parameters[section]:
+                sorted_parameters['input_data'][section][key] = value
+                flag = True
+                del sorted_parameters[key]
+                break
+        if not flag:
+            unuse_parameters[key] = value
+            del sorted_parameters[key]
+    input_data = copy.deepcopy(sorted_parameters['input_data'])
+    for key, value in input_data.items():
+        flag = False
+        if key.upper() in section_names:
+            sorted_parameters['input_data'][key.upper()] = sorted_parameters['input_data'].pop(key)
+            continue
+        for section in section_names:
+            if section.upper() == 'INPUT_NTYP':
+                continue
+            if key in pw_parameters[section]:
+                sorted_parameters['input_data'][section][key] = value
+                flag = True
+                del sorted_parameters['input_data'][key]
+                break
+        if not flag:
+            unuse_parameters[key] = value
+            del sorted_parameters['input_data'][key]
+    return sorted_parameters, unuse_parameters
+    #
+def check_qe_input(input_parameters, package = 'PW'):
+    from xespresso.input_parameters import qe_namespace, default_parameters
+    from xespresso.utils import check_type
+    pw_parameters = qe_namespace[package]
+    for section, parameters in input_parameters.items():
+        if section == 'INPUT_NTYP':
+            for key, subparas in parameters.items():
+                for value in subparas.values():
+                    check_type(key, value, pw_parameters)
+        else:
+            for key, value in parameters.items():
+                check_type(key, value, pw_parameters)
+
+def read_espresso_asei(fileobj, package = 'PW'):
+    """Parse Quantum ESPRESSO input parameters
+    """
+    with open(fileobj, 'rb') as fp:
+        atoms, parameters = pickle.load(fp)
+    if package == 'PW':
+        sorted_parameters, unuse_parameters = sort_qe_input(parameters)
+        check_qe_input(sorted_parameters['input_data'], package)
+    else:
+        sorted_parameters = parameters
+    return atoms, sorted_parameters
+def write_espresso_asei(fileobj, atoms, parameters):
+    """save Quantum ESPRESSO input parameters
+    """
+    with open(fileobj, 'wb') as fp:
+        pickle.dump([atoms, parameters], fp)
+
+def get_atomic_constraints(pwo, n_atoms):
+    """
+    Read constraints from QE output
+    """
+    constraints = []
+    with open(pwo, 'r') as f:
+        lines = f.readlines()
+        indexes = []
+        for i, line in enumerate(lines):
+            if 'ATOMIC_POSITIONS' in line:
+                indexes.append(i)
+        if len(indexes) > 0:
+            image_index = indexes[0] + 1
+            for line in lines[image_index:image_index + n_atoms + 1]:
+                split_line = line.split()
+                if len(split_line) > 4:
+                    constraint = (float(split_line[4]),
+                                  float(split_line[5]),
+                                  float(split_line[6]))
+                else:
+                    constraint = (None, None, None)
+                constraints.append(constraint)
+    constraints = get_constraint(constraints)
+    return constraints
+
+def get_constraint(constraint_idx):
+    """
+    Map constraints from QE output to FixAtoms or FixCartesian constraint
+    """
+    if not np.any(constraint_idx):
+        return None
+
+    a = [a for a, c in enumerate(constraint_idx) if np.all(c is not None)]
+    mask = [[(ic + 1) % 2 for ic in c] for c in constraint_idx
+            if np.all(c is not None)]
+
+    if np.all(np.array(mask)) == 1:
+        constraint = FixAtoms(a)
+    else:
+        constraint = FixCartesian(a, mask)
+    return constraint
